import pytest
from snakestream.collector import to_list

from snakestream.core import Stream


@pytest.mark.asyncio
async def test_builder_simple() -> None:
    # given
    builder = Stream.builder()
    builder.add(1)
    builder.add(5)
    builder.add(8)

    # when
    stream = builder.build()
    it = await stream.collect(to_list)

    # then
    assert it == [1, 5, 8]


@pytest.mark.asyncio
async def test_builder_empty() -> None:
    # given
    builder = Stream.builder()

    # when
    stream = builder.build()
    it = await stream.collect(to_list)

    # then
    assert it == []



import pytest
import asyncio

from snakestream import stream_of
from snakestream.collector import to_list


@pytest.mark.asyncio
async def test_sorted() -> None:
    outset = [1, 5, 3, 4, 5, 2]

    actual = await stream_of(outset) \
        .sorted() \
        .collect(to_list)

    assert sorted(outset) == actual


@pytest.mark.asyncio
async def test_sorted_reverse() -> None:
    outset = [1, 5, 3, 4, 5, 2]

    actual = await stream_of(outset) \
        .sorted(reverse=True) \
        .collect(to_list)

    assert sorted(outset, reverse=True) == actual


@pytest.mark.asyncio
async def test_sorted_comparator() -> None:
    outset = [
        {'x': 1, 'y': 5},
        {'x': 3, 'y': 7},
        {'x': 2, 'y': 6},
    ]

    def compare(a, b):
        if a['x'] > b['x']:
            return 1
        elif a['x'] < b['x']:
            return -1
        else:
            return 0

    actual = await stream_of(outset) \
        .sorted(comparator=compare) \
        .collect(to_list)

    assert sorted(outset, key=lambda x: x['x']) == actual


@pytest.mark.asyncio
async def test_sorted_async_comparator_and_reverse() -> None:
    outset = [
        {'x': 1, 'y': 5},
        {'x': 3, 'y': 7},
        {'x': 2, 'y': 6},
    ]

    async def compare_async(a, b):
        await asyncio.sleep(0.01)
        if a['x'] == b['x']:
            return 0
        elif a['x'] > b['x']:
            return 1
        else:
            return -1

    actual = await stream_of(outset) \
        .sorted(comparator=compare_async, reverse=True) \
        .collect(to_list)

    assert actual == [
        {'x': 3, 'y': 7},
        {'x': 2, 'y': 6},
        {'x': 1, 'y': 5},
    ]



import pytest

from snakestream import stream_of
from snakestream.collector import to_list, to_generator


@pytest.mark.asyncio
async def test_to_generator() -> None:
    # when
    it = stream_of([1, 2, 3, 4]) \
        .collect(to_generator)
    # then
    assert await it.__anext__() == 1
    assert await it.__anext__() == 2
    assert await it.__anext__() == 3
    assert await it.__anext__() == 4

    with pytest.raises(StopAsyncIteration):
        await it.__anext__()


@pytest.mark.asyncio
async def test_to_generator_with_null_in_stream() -> None:
    # when
    it = stream_of([1, 2, None, 4]) \
        .collect(to_generator)
    # then
    assert await it.__anext__() == 1
    assert await it.__anext__() == 2
    assert await it.__anext__() is None
    assert await it.__anext__() == 4

    with pytest.raises(StopAsyncIteration):
        await it.__anext__()


@pytest.mark.asyncio
async def test_to_generator_with_empty_list_input() -> None:
    # when
    it = stream_of([]) \
        .collect(to_generator)
    # then
    with pytest.raises(StopAsyncIteration):
        await it.__anext__()


@pytest.mark.asyncio
async def test_to_list() -> None:
    # when
    it = await stream_of([1, 2, 3, 4]) \
        .collect(to_list)
    # then
    assert it == [1, 2, 3, 4]


@pytest.mark.asyncio
async def test_to_list_with_none_in_stream() -> None:
    # when
    it = await stream_of([1, None, 3, 4]) \
        .collect(to_list)
    # then
    assert it == [1, None, 3, 4]


@pytest.mark.asyncio
async def test_to_list_with_empty_list_input() -> None:
    # when
    it = await stream_of([]) \
        .collect(to_list)
    # then
    assert it == []



import pytest
import asyncio

from snakestream import stream_of
from snakestream.collector import to_generator, to_list


@pytest.mark.asyncio
async def test_filter_multiple() -> None:
    # when
    it = stream_of([1, 2, 3, 4, 5, 6]) \
        .filter(lambda x: x > 3) \
        .filter(lambda x: x < 6) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 4
    assert await it.__anext__() == 5
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False


@pytest.mark.asyncio
async def test_filter_does_not_mutate_source() -> None:
    source = [1, 2, 3, 4, 5, 6]

    # when
    it = await stream_of(source) \
        .filter(lambda x: x > 3) \
        .filter(lambda x: x < 6) \
        .collect(to_list)

    # then
    assert source != it
    assert len(source) == 6
    assert len(it) == 2


@pytest.mark.asyncio
async def test_filter_async_function() -> None:

    async def async_predicate(x: int) -> bool:
        await asyncio.sleep(0.01)
        return x < 3

    # when
    it = stream_of([1, 2, 3, 4]) \
        .filter(async_predicate) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 1
    assert await it.__anext__() == 2
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name

from typing import AsyncGenerator
import pytest

from snakestream import stream_of
from snakestream.collector import to_generator, to_list


async def async_generator() -> AsyncGenerator:
    for i in range(1, 6):
        yield i


class AsyncIteratorImpl:
    def __init__(self, end_range):
        self.end = end_range
        self.start = -1

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.start < self.end - 1:
            self.start += 1
            return self.start
        raise StopAsyncIteration


@pytest.mark.asyncio
async def test_input_list() -> None:
    # when
    it = stream_of([1, 2, 3, 4]) \
        .collect(to_generator)
    # then
    assert await it.__anext__() == 1
    assert await it.__anext__() == 2
    assert await it.__anext__() == 3
    assert await it.__anext__() == 4
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False


@pytest.mark.asyncio
async def test_input_async_generator() -> None:
    # when
    it = stream_of(async_generator()) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 1
    assert await it.__anext__() == 2
    assert await it.__anext__() == 3
    assert await it.__anext__() == 4
    assert await it.__anext__() == 5
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False


@pytest.mark.asyncio
async def test_input_async_iterator() -> None:
    # when
    it = stream_of(AsyncIteratorImpl(5)) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 0
    assert await it.__anext__() == 1
    assert await it.__anext__() == 2
    assert await it.__anext__() == 3
    assert await it.__anext__() == 4
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False


@pytest.mark.asyncio
async def test_null_input() -> None:
    # when
    with pytest.raises(TypeError):
        await stream_of(None) \
            .collect(to_list)



import pytest

from snakestream import stream_of
from snakestream.collector import to_list
from conftest import MyObject


@pytest.mark.asyncio
async def test_unique() -> None:
    # when
    it = await stream_of([1, 7, 3, 7, 5, 6, 0, 6, 6]) \
        .distinct() \
        .collect(to_list)
    # then
    assert it == [1, 7, 3, 5, 6, 0]


@pytest.mark.asyncio
async def test_unique_empty_list() -> None:
    # when
    it = await stream_of([]) \
        .distinct() \
        .collect(to_list)
    # then
    assert it == []


@pytest.mark.asyncio
async def test_unique_list_with_no_dupes() -> None:
    # when
    it = await stream_of([1, 2, 3, 4]) \
        .distinct() \
        .collect(to_list)
    # then
    assert it == [1, 2, 3, 4]


@pytest.mark.asyncio
async def test_unique_object_list() -> None:
    # when
    input_list = [MyObject(1, "object1"), MyObject(2, "object2"), MyObject(3, "object3"), MyObject(2, "object2"),
                  MyObject(3, "object3")]
    it = await stream_of(input_list) \
        .distinct() \
        .collect(to_list)
    # then
    assert it == [MyObject(1, "object1"), MyObject(2, "object2"), MyObject(3, "object3")]



import pytest
import asyncio

from snakestream import stream_of


coords = [
    {'x': 1, 'y': 5},
    {'x': 2, 'y': 6},
    {'x': 3, 'y': 7},
]


@pytest.mark.asyncio
async def test_for_each() -> None:
    def incr_y(c) -> None:
        c['y'] = 1

    await stream_of(coords) \
        .for_each(incr_y)

    assert coords[0]['y'] == 1
    assert coords[1]['y'] == 1
    assert coords[2]['y'] == 1


@pytest.mark.asyncio
async def test_for_each_async() -> None:
    async def async_incr_y(c) -> None:
        await asyncio.sleep(0.01)
        c['y'] = 1

    await stream_of(coords) \
        .for_each(async_incr_y)

    assert coords[0]['y'] == 1
    assert coords[1]['y'] == 1
    assert coords[2]['y'] == 1



import pytest

from snakestream import stream_of


@pytest.mark.asyncio
async def test_empty_stream() -> None:
    it = await stream_of([]) \
        .count()
    assert it == 0


@pytest.mark.asyncio
async def test_base_test() -> None:
    it = await stream_of([1, 2, 3, 4, 5, 6]) \
        .count()
    assert it == 6


@pytest.mark.asyncio
async def test_case_insensitive() -> None:
    it = await stream_of(['test', 'Test', 'test', 'foo', 'bar']) \
        .count()
    assert it == 5


@pytest.mark.asyncio
async def test_null_count() -> None:
    it = await stream_of(['test', 'Test', 'test', None, 'bar']) \
        .count()
    assert it == 5



import asyncio

import pytest

from snakestream import stream_of


@pytest.mark.asyncio
async def test_reducer() -> None:
    # when
    it = stream_of([1, 2, 3, 4, 5, 6]) \
        .reduce(0, lambda x, y: x + y)
    # then
    assert await it == 21


@pytest.mark.asyncio
async def test_reducer_associative() -> None:
    # when
    it = stream_of([1, 2, 3, 4, 5, 6]) \
        .reduce(0, lambda x, y: x + y)

    it2 = stream_of([1, 2, 3, 4, 5, 6]) \
        .reduce(0, lambda x, y: y + x)
    # then
    assert await it == 21
    assert await it2 == 21


@pytest.mark.asyncio
async def test_async_reducer() -> None:
    async def async_reducer(x: int, y: int):
        await asyncio.sleep(0.01)
        return x + y

    # when
    it = stream_of([1, 2, 3, 4, 5, 6]) \
        .reduce(0, async_reducer)

    # then
    assert await it == 21


@pytest.mark.asyncio
async def test_reducer_mixed_chain(letter_2_int) -> None:
    # when
    it = stream_of(['a', 'b', 'c', 'd']) \
        .map(lambda x: letter_2_int[x]) \
        .reduce(0, lambda x, y: x + y)
    # then
    assert await it == 10




from typing import List

from snakestream.core import Stream
from snakestream.type import T


class StreamBuilder:
    def __init__(self) -> None:
        self._elements: List[T] = []

    def add(self, element: T) -> 'StreamBuilder':
        self.accept(element)
        return self

    def accept(self, element: T) -> None:
        self._elements.append(element)

    def build(self) -> 'Stream':
        return Stream(self._elements)



import pytest

from snakestream import stream_of, Stream
from snakestream.collector import to_list


@pytest.mark.asyncio
async def test_empty_stream() -> None:
    # when
    expected = stream_of([])
    actual = Stream.empty()

    assert await expected.collect(to_list) == await actual.collect(to_list)



import abc
from typing import Any, AsyncGenerator, AsyncIterable, Awaitable, Callable, Generator, Iterable, Optional, TypeVar, Union


#
# Generic types
#
T = TypeVar('T')
R = TypeVar('R')

Streamable = Union[Iterable, AsyncIterable, Generator, AsyncGenerator]

Predicate = Callable[[T], Union[bool, Awaitable[bool]]]

# Intermediaries
Filterer = Callable[[T], T]
Mapper = Callable[[T], Optional[R]]
FlatMapper = Callable[[Streamable], 'AbstractStream']
Comparator = Callable[[T, T], Union[bool, Awaitable[bool]]]
Consumer = Callable[[T], T]

# Terminals
Accumulator = Callable[[T, Union[T, R]], Union[T, R]]


#
# Classes
#
class AbstractStream(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractclassmethod
    def empty() -> 'AbstractStream':
        raise NotImplementedError

    @staticmethod
    @abc.abstractclassmethod
    async def concat(a: 'AbstractStream', b: 'AbstractStream') -> 'AbstractStream':
        raise NotImplementedError

    @staticmethod
    @abc.abstractclassmethod
    def builder() -> 'AbstractStreamBuilder':
        raise NotImplementedError

    # Intermediaries
    @abc.abstractclassmethod
    def filter(self, predicate: Predicate) -> 'AbstractStream':
        raise NotImplementedError

    @abc.abstractclassmethod
    def map(self, mapper: Mapper) -> 'AbstractStream':
        raise NotImplementedError

    @abc.abstractclassmethod
    def flat_map(self, flat_mapper: FlatMapper) -> 'AbstractStream':
        raise NotImplementedError

    @abc.abstractclassmethod
    def sorted(self, comparator: Optional[Comparator] = None, reverse=False) -> 'AbstractStream':
        raise NotImplementedError

    @abc.abstractclassmethod
    def distinct(self) -> 'AbstractStream':
        raise NotImplementedError

    @abc.abstractclassmethod
    def peek(self, consumer: Consumer) -> 'AbstractStream':
        raise NotImplementedError

    # Terminals
    @abc.abstractclassmethod
    def collect(self, collector: Callable) -> AsyncGenerator:
        raise NotImplementedError

    @abc.abstractclassmethod
    async def reduce(self, identity: Union[T, R], accumulator: Accumulator) -> Union[T, R]:
        raise NotImplementedError

    @abc.abstractclassmethod
    async def for_each(self, consumer: Callable[[T], Any]) -> None:
        raise NotImplementedError

    @abc.abstractclassmethod
    async def find_first(self) -> Optional[Any]:
        raise NotImplementedError

    @abc.abstractclassmethod
    async def max(self, comparator: Comparator) -> Optional[T]:
        raise NotImplementedError

    @abc.abstractclassmethod
    async def min(self, comparator: Comparator) -> Optional[T]:
        raise NotImplementedError

    @abc.abstractclassmethod
    async def all_match(self, predicate: Predicate) -> bool:
        raise NotImplementedError

    @abc.abstractclassmethod
    async def none_match(self, predicate: Predicate) -> bool:
        raise NotImplementedError

    @abc.abstractclassmethod
    async def any_match(self, predicate: Predicate) -> bool:
        raise NotImplementedError

    @abc.abstractclassmethod
    async def count(self) -> int:
        raise NotImplementedError


class AbstractStreamBuilder(metaclass=abc.ABCMeta):

    @abc.abstractclassmethod
    def add(self, element: T) -> 'AbstractStreamBuilder':
        raise NotImplementedError

    @abc.abstractclassmethod
    def accept(self, element: T) -> None:
        raise NotImplementedError

    @abc.abstractclassmethod
    def build(self) -> 'AbstractStream':
        raise NotImplementedError
